/* MathLinkWrapper.c */

/* $Id$ */

/* This file is the interface between MathLinkInterface.f90 and the
 * actual MathLink functions. It calls the different MLPut and
 * MLGet functions, and it contains a main function interface which
 * is called from the PROGRAM part in the MathCode generated
 * Fortran90 code. 
 *
 * The first thing that will strike you is the fact that all functions
 * that are to be called from Fortran90 have different names on
 * different architectures. A function declared as "void W_STDCALL FOO"
 * on Windows will be "void foo_" on Unix. This has to do with the fact
 * that different Fortran90 compilers generate different external
 * names. Trust me, this is the best solution.
 
 
 * Note use of STDCALL macro. It is needed to be in DF60,
 * but cannot (!) be in IFORT80  
 
 * IN DF60/MSVC.NET:
 * When in Fortran you write call A(7), then it looks external symbol _A@4
 * then in C this function must NOT be declared as
 * void cdecl A(int x) and it becomes external symbol _A as well.
 * "cdecl" is  default.
 * If howerer one writes
 * void __stdcall A(int x), it becomes the external symbol _A@4  
 * Code must be complied with DF60 flag !
 
 * IN IFORT80/VS.NET:
 * When in Fortran you write call A(7), then it looks external symbol _A
 * then in C this function must be declared as
 * void cdecl A(int x) and it becomes external symbol _A as well.
 * "cdecl" is  default.
 * If howerer one writes
 * void __stdcall A(int x), it becomes the external symbol _A@4  
 * Code must be complied with IFORT80 flag !
   
 */

/* Here is an example of how it might look with a package foo
 * containing a function bar that takes an integer vector and
 * returns an integer vector.
 

   MathCodeMainUnx.c           MathLinkWrapper.c             footm.c
+--------------------+    +--------------------+    +--------------------+
|                    |    |                    |    |                    |
|     PROGRAM  --------------->                |    |   mprep generated  |
|                    |    |                    |    |       C code       |
|                    |  +---->                 |    |                    |
+--------------------+  | |     MLIF_MAIN -------------> MLMainString    |
                        | |                    |    |         |          |
  MathLinkWrapper.c     | |                    |    |        \|/         |
+--------------------+  | |       RECEIVENINT  |    |  waiting for call  |
| (in G95 variant)   |  | |             /|\    |    |  from Mathematica  |
|  main()           ----+ |   SENDNINT   |     |    |         |          |
|                    |    |     /|\      |     |    |         |          |
+--------------------+    +------|-------|-----+    +---------|----------+
                                 |       |                    |
+--------------------+    +------|-------|-----+    +---------|----------+
|                    |    |      |       |     |    |         |          |
|  MathCode generated|    |      |       |     |    |        \|/         |
|    Fortran90 code  |    |      |       |     |    |       ML_BAR       |
|                    |    |      |       |     |    |         |          |
|                    |    |      |    getNint <---------------+          |
|                    |    |      |             |    |         |          |
|     foo_Tbar <-------.. |      |             | ..-----------+          |
|                    |    |      |             |    |         |          |
|                    |    |   putNint <-----------------------+          |
|                    |    |                    |    |                    |
+--------------------+    +--------------------+    +--------------------+
      foo.f90              MathLinkInterface.f90          fooif.f90
      
      
      
 */


#ifdef df60
#define W_STDCALL __stdcall
#define CAPITAL
#else

#ifdef df50
#define W_STDCALL __stdcall
#define CAPITAL
#else

#ifdef ifort80
#define W_STDCALL 
#define CAPITAL
#else

#ifdef ifort808
#define W_STDCALL 
#define CAPITAL
#else

#ifdef absoft
#define W_STDCALL
#define CAPITAL 
#else

#ifdef  g95 
#define W_STDCALL
#define CAPITAL  
#else

#ifdef  gfortran 
#define PREFIX
#else

#ifdef  ugfortran 
// Underscore at the end
#else


#ifdef  u95 
#define W_STDCALL
#else

#ifdef  uif 
#define W_STDCALL
#else

#ifdef  unix_style
#define W_STDCALL 
#else

 See above to understand which flag to use !
#endif
#endif
#endif
#endif
#endif
#endif
#endif
#endif
#endif
#endif
#endif

#ifndef __APPLE__
/* in Darwin this file is in malloc/malloc.h ,
    but malloc() is in stdlib.h anyway */ 
#include <malloc.h>   /* malloc() */
#endif

#include <string.h>   /* memcpy() */

#include <stdlib.h>   /* exit(int) */

#include <stdio.h>    /* stderr */

#include "mathlink.h"


/* In mathlink.h several functions changed their argument type when going over to MLINTERFACE==3  */
#if MLINTERFACE >=3
    #define int3long int 
#else
    #define int3long long 
#endif


/* Global variables for transferring data */
int *local_data_int;
double *local_data_double;
int *local_dims;
char **local_heads;
int local_depth;
long output_dims[4];

/* Global variables for callback */
MLENV cb_env = (MLENV)0;
int cb_next;
char cb_current_function[100];
int cb_error;

/* Prototypes for callback */
static void init_and_openlink(int argc, char *argv[]);
static int read_and_print_expression(int next);
void MLIF_error();
void MLIF_expected(char *typetext);
#ifdef CAPITAL
void W_STDCALL MLIF_START(int *argc, char *fname, unsigned int fnamelen);
void W_STDCALL MLIF_WAIT();
#else
void mlif_start_(int *argc, char *fname, unsigned int fnamelen);
void mlif_wait_();
#endif

/* ------------------------------------------------------------------
 *  Main function called from the Fortran90 PROGRAM part
 */

/* Note: This is a C function called from Fortran90 that receives
 * a string. This is tricky. For DIGITAL Visual Fortran, the
 * compilation option /iface:mixed_str_len_arg is default.
 * This means that the string length is passed immediately after
 * the string itself and not at the end. (In this case it is ok since
 * we only have one variable, but below in the function MLIF_START
 * things get a little more complicated.)
 */

#ifdef CAPITAL
void W_STDCALL MLIF_MAIN(char *cmdline, unsigned int cmdlinelen)
#else
void mlif_main_(char *cmdline, unsigned int cmdlinelen)
#endif
{
	char *cmdline2 = (char *)malloc(cmdlinelen+1);
	memcpy((void *)cmdline2, (void *)cmdline, cmdlinelen);
	*(cmdline2+cmdlinelen) = '\0';
	MLMainString(cmdline2);
}

/* Same as above but adjusted to g95 compiler rules */
#if 0
// EXPERIMENTAL CODE, NOT IN USE

#ifdef CAPITAL
void W_STDCALL MLIF_MAIN_(char *cmdline, unsigned int cmdlinelen)
#else
void mlif_main__(char *cmdline, unsigned int cmdlinelen)
#endif
{
extern  void g95_runtime_start(int argc, char* argv[]); 
extern  void g95_runtime_stop(); 

	char *cmdline2 = (char *)malloc(cmdlinelen+1);
	memcpy((void *)cmdline2, (void *)cmdline, cmdlinelen);
	*(cmdline2+cmdlinelen) = '\0';
printf("Bef 	g95_runtime_start \n");
	/*g95_runtime_start(0,0);*/
printf("Aft 	g95_runtime_start \n");
	
	MLMainString(cmdline2);
	/*g95_runtime_stop();*/
}
#endif


#ifdef g95
extern int MLMain( int argc, char** argv);

int main(int argc, char ** argv) {
        return MLMain(argc, argv);
}
#endif

#ifdef gfortran
extern int MLMain( int argc, char** argv);

int main(int argc, char ** argv) {
        return MLMain(argc, argv);
}
#endif

#ifdef ugfortran
extern int MLMain( int argc, char** argv);

int main(int argc, char ** argv) {
        return MLMain(argc, argv);
}
#endif



#ifdef u95
extern int MLMain( int argc, char** argv);

extern char **f__xargv;
extern int f__xargc;

int MAIN_() {
    // Fix in MathCode F90 1.2.3 : MAIN_ has no arguments!
    return MLMain(f__xargc, f__xargv);
}
#endif



#ifdef uif
extern int MLMain( int argc, char** argv);
    extern int for__l_argc;
    extern char **for__a_argv;

int MAIN__() {
    //printf("argc=%d\n",argc);
        printf("argc=%d\n",for__l_argc);
        return MLMain(for__l_argc, for__a_argv);
}
#endif

/* ------------------------------------------------------------------
 *  Functions used for prefixing return of multiple values
 */

#ifdef CAPITAL
void W_STDCALL MLIF_LIST(int *count)
#else
#ifdef PREFIX
void mlif_list(int *count)
#else
void mlif_list_(int *count)
#endif
#endif

{
	MLPutFunction(stdlink, "List", *count);
}

/* ------------------------------------------------------------------
 *  Transfer simple variables
 */

#ifdef CAPITAL
void W_STDCALL SENDNULL()
#else
#ifdef PREFIX
void sendnull()
#else
void sendnull_()
#endif
#endif

{
    MLPutSymbol(stdlink, "Null");
}

#ifdef CAPITAL
void W_STDCALL RECEIVEINT(int *i)
#else
#ifdef PREFIX
void receiveint(int *i)
#else
void receiveint_(int *i)
#endif
#endif

{
	MLGetInteger(stdlink, i);
}

#ifdef CAPITAL
void W_STDCALL SENDINT(int *i)
#else
#ifdef PREFIX
void sendint(int *i)
#else
void sendint_(int *i)
#endif
#endif

{
	MLPutInteger(stdlink, *i);
}

#ifdef CAPITAL
void W_STDCALL SENDSTR(char *s)
#else
#ifdef PREFIX
void sendstr(char* s)
#else
void sendstr_(char* s)
#endif
#endif
{
    s[255]=0;
    MLPutString(stdlink, s);
}



#ifdef CAPITAL
void W_STDCALL RECEIVEDOUBLE(double *d)
#else
#ifdef PREFIX
void receivedouble(double *d)
#else
void receivedouble_(double *d)
#endif
#endif

{
	MLGetDouble(stdlink, d);
}

#ifdef CAPITAL
void W_STDCALL SENDDOUBLE(double *d)
#else
#ifdef PREFIX
void senddouble(double *d)
#else
void senddouble_(double *d)
#endif
#endif

{
	MLPutDouble(stdlink, *d);
}


#ifdef CAPITAL
void W_STDCALL RECEIVESTR(char *s,int * len)
#else
#ifdef PREFIX
void receivestr(char *s, int * len)
#else
    void receivestr_(char *s, int * len)
#endif
#endif
{
    const char * string;
    MLGetString(stdlink,&string); 
    strcpy(s,string);
    *len=strlen(string);
    fprintf(stderr,"s=(%s) len=(%d)\n",s,*len);
}



/* TODO: Write receive and send for strings and vectors of strings */

/* ------------------------------------------------------------------
 *  Transfer vectors
 */

#ifdef CAPITAL
void W_STDCALL RECEIVENINT(int *remote_depth)
#else
#ifdef PREFIX
void receivenint(int *remote_depth)
#else
void receivenint_(int *remote_depth)
#endif
#endif

{
    MLGetInteger32List(stdlink, &local_data_int, &local_depth);
	*remote_depth = (int)local_depth;
}

#ifdef CAPITAL
void W_STDCALL COPYNINT(int *remote_data_int)
#else
#ifdef PREFIX
void copynint(int *remote_data_int)
#else
void copynint_(int *remote_data_int)
#endif
#endif

{
	memcpy((void *)remote_data_int, (void *)local_data_int,
	       sizeof(int)*local_depth);
    MLReleaseInteger32List(stdlink, local_data_int, local_depth);
}

#ifdef CAPITAL
void W_STDCALL SENDNINT(int *remote_data_int, int *remote_depth)
#else
#ifdef PREFIX
void sendnint(int *remote_data_int, int *remote_depth)
#else
void sendnint_(int *remote_data_int, int *remote_depth)
#endif
#endif

{
	MLPutIntegerList(stdlink, remote_data_int, (long)*remote_depth);
}




#ifdef CAPITAL
void W_STDCALL RECEIVENDOUBLE(int *remote_depth)
#else
#ifdef PREFIX
void receivendouble(int *remote_depth)
#else
void receivendouble_(int *remote_depth)
#endif
#endif
{
    MLGetReal64List(stdlink, &local_data_double, &local_depth);
	*remote_depth = (int)local_depth;
}

#ifdef CAPITAL
void W_STDCALL COPYNDOUBLE(double *remote_data_double)
#else
#ifdef PREFIX
void copyndouble(double *remote_data_double)
#else
void copyndouble_(double *remote_data_double)
#endif
#endif
{
	memcpy((void *)remote_data_double, (void *)local_data_double, sizeof(double)*local_depth);
	MLReleaseReal64List(stdlink, local_data_double, local_depth);
}

#ifdef CAPITAL
void W_STDCALL SENDNDOUBLE(double *remote_data_double, int *remote_depth)
#else
#ifdef PREFIX
void sendndouble(double *remote_data_double, int *remote_depth)
#else
void sendndouble_(double *remote_data_double, int *remote_depth)
#endif
#endif
{
	MLPutRealList(stdlink, remote_data_double, (long)*remote_depth);
}

/* ------------------------------------------------------------------
 *  Transfer 2D arrays
 */

#ifdef CAPITAL
void W_STDCALL RECEIVENNINT(int *remote_dims)
#else
#ifdef PREFIX
void receivennint(int *remote_dims)
#else
void receivennint_(int *remote_dims)
#endif
#endif
{
    MLGetInteger32Array(stdlink, &local_data_int, &local_dims,
	                  &local_heads, &local_depth);
	*remote_dims = (int)local_dims[0];
	*(remote_dims+1) = (int)local_dims[1];
}

#ifdef CAPITAL
void W_STDCALL COPYNNINT(int *remote_data_int)
#else
#ifdef PREFIX
void copynnint(int *remote_data_int)
#else
void copynnint_(int *remote_data_int)
#endif
#endif
{
	memcpy((void *)remote_data_int, (void *)local_data_int,
	       sizeof(int)*local_dims[0]*local_dims[1]);
    MLReleaseInteger32Array(stdlink, local_data_int, local_dims,
	                     local_heads, local_depth);
}

#ifdef CAPITAL
void W_STDCALL SENDNNINT(int *remote_data_int, int *remote_dims)
#else
#ifdef PREFIX
void sendnnint(int *remote_data_int, int *remote_dims)
#else
void sendnnint_(int *remote_data_int, int *remote_dims)
#endif
#endif
{
	output_dims[0] = *remote_dims;
	output_dims[1] = *(remote_dims+1);
	MLPutIntegerArray(stdlink, remote_data_int, output_dims, NULL, 2);
}


#ifdef CAPITAL
void W_STDCALL RECEIVENNDOUBLE(int *remote_dims)
#else
#ifdef PREFIX
void receivenndouble(int *remote_dims)
#else
void receivenndouble_(int *remote_dims)
#endif
#endif
{
    MLGetReal64Array(stdlink, &local_data_double, &local_dims,
	               &local_heads, &local_depth);
	*remote_dims = (int)local_dims[0];
	*(remote_dims+1) = (int)local_dims[1];
}

#ifdef CAPITAL
void W_STDCALL COPYNNDOUBLE(double *remote_data_double)
#else
#ifdef PREFIX
void copynndouble(double *remote_data_double)
#else
void copynndouble_(double *remote_data_double)
#endif
#endif
{
	memcpy((void *)remote_data_double, (void *)local_data_double,
	       sizeof(double)*local_dims[0]*local_dims[1]);
     MLReleaseReal64Array(stdlink, local_data_double, local_dims,
	                  local_heads, local_depth);
}

#ifdef CAPITAL
void W_STDCALL SENDNNDOUBLE(double *remote_data_double, int *remote_dims)
#else
#ifdef PREFIX
void sendnndouble(double *remote_data_double, int *remote_dims)
#else
void sendnndouble_(double *remote_data_double, int *remote_dims)
#endif
#endif
{
	output_dims[0] = *remote_dims;
	output_dims[1] = *(remote_dims+1);
	MLPutRealArray(stdlink, remote_data_double, output_dims, NULL, 2);
}

/* ------------------------------------------------------------------
 *  Transfer 3D arrays
 */

#ifdef CAPITAL
void W_STDCALL RECEIVENNNINT(int *remote_dims)
#else
#ifdef PREFIX
void receivennnint(int *remote_dims)
#else
void receivennnint_(int *remote_dims)
#endif
#endif

{
    MLGetInteger32Array(stdlink, &local_data_int, &local_dims,
	                  &local_heads, &local_depth);
	*remote_dims = (int)local_dims[0];
	*(remote_dims+1) = (int)local_dims[1];
	*(remote_dims+2) = (int)local_dims[2];
}

#ifdef CAPITAL
void W_STDCALL COPYNNNINT(int *remote_data_int)
#else
#ifdef PREFIX
void copynnnint(int *remote_data_int)
#else
void copynnnint_(int *remote_data_int)
#endif
#endif

{
	memcpy((void *)remote_data_int, (void *)local_data_int,
	       sizeof(int)*local_dims[0]*local_dims[1]*local_dims[2]);
	MLReleaseInteger32Array(stdlink, local_data_int, local_dims,
	                     local_heads, local_depth);
}

#ifdef CAPITAL
void W_STDCALL SENDNNNINT(int *remote_data_int, int *remote_dims)
#else
#ifdef PREFIX
void sendnnnint(int *remote_data_int, int *remote_dims)
#else
void sendnnnint_(int *remote_data_int, int *remote_dims)
#endif
#endif

{
	output_dims[0] = *remote_dims;
	output_dims[1] = *(remote_dims+1);
	output_dims[2] = *(remote_dims+2);
	MLPutIntegerArray(stdlink, remote_data_int, output_dims, NULL, 3);
}


#ifdef CAPITAL
void W_STDCALL RECEIVENNNDOUBLE(int *remote_dims)
#else
#ifdef PREFIX
void receivennndouble(int *remote_dims)
#else
void receivennndouble_(int *remote_dims)
#endif
#endif
{
    MLGetReal64Array(stdlink, &local_data_double, &local_dims,
	               &local_heads, &local_depth);
	*remote_dims = (int)local_dims[0];
	*(remote_dims+1) = (int)local_dims[1];
	*(remote_dims+2) = (int)local_dims[2];
}

#ifdef CAPITAL
void W_STDCALL COPYNNNDOUBLE(double *remote_data_double)
#else
#ifdef PREFIX
void copynnndouble(double *remote_data_double)
#else
void copynnndouble_(double *remote_data_double)
#endif
#endif
{
	memcpy((void *)remote_data_double, (void *)local_data_double,
	       sizeof(double)*local_dims[0]*local_dims[1]*local_dims[2]);
    MLReleaseReal64Array(stdlink, local_data_double, local_dims,
	                  local_heads, local_depth);
}

#ifdef CAPITAL
void W_STDCALL SENDNNNDOUBLE(double *remote_data_double, int *remote_dims)
#else
#ifdef PREFIX
void sendnnndouble(double *remote_data_double, int *remote_dims)
#else
void sendnnndouble_(double *remote_data_double, int *remote_dims)
#endif
#endif
{
	output_dims[0] = *remote_dims;
	output_dims[1] = *(remote_dims+1);
	output_dims[2] = *(remote_dims+2);
	MLPutRealArray(stdlink, remote_data_double, output_dims, NULL, 3);
}

/* ------------------------------------------------------------------
 *  Transfer 4D arrays
 */

#ifdef CAPITAL
void W_STDCALL RECEIVENNNNINT(int *remote_dims)
#else
#ifdef PREFIX
void receivennnnint(int *remote_dims)
#else
void receivennnnint_(int *remote_dims)
#endif
#endif

{
    MLGetInteger32Array(stdlink, &local_data_int, &local_dims,
	                  &local_heads, &local_depth);
	*remote_dims = (int)local_dims[0];
	*(remote_dims+1) = (int)local_dims[1];
	*(remote_dims+2) = (int)local_dims[2];
	*(remote_dims+3) = (int)local_dims[3];
}

#ifdef CAPITAL
void W_STDCALL COPYNNNNINT(int *remote_data_int)
#else
#ifdef PREFIX
void copynnnnint(int *remote_data_int)
#else
void copynnnnint_(int *remote_data_int)
#endif
#endif
{
	memcpy((void *)remote_data_int, (void *)local_data_int,
	       sizeof(int)*local_dims[0]*local_dims[1]*local_dims[2]*local_dims[3]);
	MLReleaseInteger32Array(stdlink, local_data_int, local_dims,
	                     local_heads, local_depth);
}

#ifdef CAPITAL
void W_STDCALL SENDNNNNINT(int *remote_data_int, int *remote_dims)
#else
#ifdef PREFIX
void sendnnnnint(int *remote_data_int, int *remote_dims)
#else
void sendnnnnint_(int *remote_data_int, int *remote_dims)
#endif
#endif
{
	output_dims[0] = *remote_dims;
	output_dims[1] = *(remote_dims+1);
	output_dims[2] = *(remote_dims+2);
	output_dims[3] = *(remote_dims+3);
	MLPutIntegerArray(stdlink, remote_data_int, output_dims, NULL, 4);
}


#ifdef CAPITAL
void W_STDCALL RECEIVENNNNDOUBLE(int *remote_dims)
#else
#ifdef PREFIX
void receivennnndouble(int *remote_dims)
#else
void receivennnndouble_(int *remote_dims)
#endif
#endif
{
    MLGetReal64Array(stdlink, &local_data_double, &local_dims,
	               &local_heads, &local_depth);
	*remote_dims = (int)local_dims[0];
	*(remote_dims+1) = (int)local_dims[1];
	*(remote_dims+2) = (int)local_dims[2];
	*(remote_dims+3) = (int)local_dims[3];
}

#ifdef CAPITAL
void W_STDCALL COPYNNNNDOUBLE(double *remote_data_double)
#else
#ifdef PREFIX
void copynnnndouble(double *remote_data_double)
#else
void copynnnndouble_(double *remote_data_double)
#endif
#endif
{
	memcpy((void *)remote_data_double, (void *)local_data_double,
	       sizeof(double)*local_dims[0]*local_dims[1]*local_dims[2]*local_dims[3]);
    MLReleaseReal64Array(stdlink, local_data_double, local_dims,
	                  local_heads, local_depth);
}

#ifdef CAPITAL
void W_STDCALL SENDNNNNDOUBLE(double *remote_data_double, int *remote_dims)
#else
#ifdef PREFIX
void sendnnnndouble(double *remote_data_double, int *remote_dims)
#else
void sendnnnndouble_(double *remote_data_double, int *remote_dims)
#endif
#endif
{
	output_dims[0] = *remote_dims;
	output_dims[1] = *(remote_dims+1);
	output_dims[2] = *(remote_dims+2);
	output_dims[3] = *(remote_dims+3);
	MLPutRealArray(stdlink, remote_data_double, output_dims, NULL, 4);
}

/* ------------------------------------------------------------------
 *  General functions used for callback
 */

/* These functions are copies of callback functions found in
 * mathinter.h in the LightMat catalog with which this file
 * should be syncronized now and then.
 */

#if defined(MINGW)&&(__GNUC__ < 4)
static void deinit(...)
#else
static void deinit(void)
#endif
{
	if (cb_env) MLDeinitialize(cb_env);
}
 
#if defined(MINGW)&&(__GNUC__ < 4)
static void closelink(...)
#else
static void closelink(void)
#endif
{
	if (stdlink) MLClose(stdlink);
}

static void init_and_openlink(int argc, char* argv[])
{
	int3long err;
	cb_env = MLInitialize((MLParametersPointer)0);
	if (cb_env == (MLENV)0) exit(1);
	atexit(deinit);
	stdlink = MLOpenArgv(cb_env, argv, argv+argc, &err);
	if (stdlink == (MLINK)0) exit(2);
	atexit(closelink);
}

static int read_and_print_expression(int next)
{
	kcharp_ct s;
	int n;
	long i;
	int3long len;
	double r;
	static int indent;
	
	switch(next) {
	case MLTKSYM:
		fprintf(stderr," SYMBOL= ");
		MLGetSymbol(stdlink, &s);
		fprintf(stderr, "%s \n", s);
        MLReleaseSymbol(stdlink, s);
		break;
	case MLTKSTR:
		fprintf(stderr," STRING=");
		MLGetString(stdlink, &s);
		fprintf(stderr, "\"%s\"\n", s);
        MLReleaseString(stdlink, s);
		break;
	case MLTKINT:
		fprintf(stderr," INT= ");
		MLGetInteger(stdlink, &n);
		fprintf(stderr, "%d \n", n);
		return n; 
		break;
	case MLTKREAL:
		fprintf(stderr," REAL= ");
		MLGetReal(stdlink, &r);
		fprintf(stderr, "%g \n", r);
		break;
	case MLTKFUNC:
		fprintf(stderr," FUNC= \n");
		indent += 3;
		fprintf(stderr, "\n %*.*s", indent, indent, "");
		if(MLGetArgCount(stdlink, &len) == 0) {
			MLIF_error();
		}
		else {
			int next2 = MLGetNext(stdlink);
			read_and_print_expression(next2);
			fprintf(stderr, "[");
			for (i = 1; i <= len; ++i){
				next2 = MLGetNext(stdlink);
				read_and_print_expression(next2);
				if (i != len) fprintf(stderr, ", ");
			}
			fprintf(stderr, "]");
		}
		indent -= 3;
		break;
	case MLTKERROR:
		fprintf(stderr, " ERROR \n");
	default:
		fprintf(stderr, " UNKNOWN Mma TYPE=%d\n", next);
		MLIF_error();
	}
	return -1;
}

void MLIF_error()
{
	if (MLError(stdlink)) {
		fprintf(stderr, "Error detected by Mathlink: %s in call to %s .\n",
			MLErrorMessage(stdlink),
			cb_current_function);
	}
	else {
		fprintf(stderr, "Error detected by this program in call to %s \n",
			cb_current_function);
	}
}

void MLIF_expected(char *typetext)
{
	fprintf(stderr, "Returning from callback function %s.\n", cb_current_function);
	fprintf(stderr, "  Expected %s, received:\n", typetext);
	read_and_print_expression(cb_next);
	fprintf(stderr, " Returns (-998) ");
}

/* ------------------------------------------------------------------
 *  Global functions used for callback (called from interface code)
 */

/* This function starts to call a callback function.
 * We also store the name of the callback function for future
 * error messages.
 *
 * (See note on passing of Fortran90 strings at MLIF_MAIN.)
 */
#ifdef CAPITAL
void W_STDCALL MLIF_START(int *argc, char *fname, unsigned int fnamelen)
#else
void mlif_start_(int *argc, char *fname, unsigned int fnamelen)
#endif
{
	char *fname2 = (char *)malloc(fnamelen+1);
	memcpy((void *)fname2, (void *)fname, fnamelen);
	*(fname2+fnamelen) = '\0';

	printf("MLIF_START: Calling function '%s' (%d chars) with %d arguments.\n", fname2, fnamelen, *argc);
	MLPutFunction(stdlink, "EvaluatePacket", 1L);
	MLPutFunction(stdlink, fname2, *argc);
	strncpy(cb_current_function, fname, 99);
	cb_current_function[99] = '\0';
}

/* This function finishes the call and waits for the result.
 */
#ifdef CAPITAL
void W_STDCALL MLIF_WAIT()
#else
void mlif_wait_()
#endif
{
	int pkt;
	MLEndPacket(stdlink);
	while((pkt = MLNextPacket(stdlink), pkt) && pkt != RETURNPKT) {
		MLNewPacket(stdlink);
		if(MLError(stdlink)) MLIF_error();
	}
}

/* ------------------------------------------------------------------
 *  Transfer simple variables (callback)
 */

/* These are special RECEIVE and COPY functions with a little
 * more error check. They use the same variables though.
 */

#ifdef CAPITAL
void W_STDCALL CBRECEIVENULL()
#else
#ifdef PREFIX
void cbreceivenull()
#else
void cbreceivenull_()
#endif
#endif
{
	cb_next = MLGetNext(stdlink);
	if (cb_next != MLTKSYM) {
		MLIF_expected("Null");
	}
	else {
		kcharp_ct s;
		MLGetSymbol(stdlink, &s);
		if (MLError(stdlink)) {
			MLIF_error();
			MLClearError(stdlink);
		}
		MLReleaseSymbol(stdlink, s);
	}
}

#ifdef CAPITAL
void W_STDCALL CBRECEIVEINT(int *i)
#else
#ifdef PREFIX
void cbreceiveint(int *i)
#else
void cbreceiveint_(int *i)
#endif
#endif
{
	cb_next = MLGetNext(stdlink);
	if (cb_next != MLTKINT) {
		MLIF_expected("Integer");
		*i = -998;
	}
	else {
		MLGetInteger(stdlink, i);
		if (MLError(stdlink)) { 
			MLIF_error();
			*i = -997;
			MLClearError(stdlink);
		}
	}
}

#ifdef CAPITAL
void W_STDCALL CBRECEIVEDOUBLE(double *d)
#else
#ifdef PREFIX
void cbreceivedouble(double *d)
#else
void cbreceivedouble_(double *d)
#endif
#endif
{
	cb_next = MLGetNext(stdlink);
	if (cb_next != MLTKREAL) {
		MLIF_expected("Real");
		*d = -998;
	}
	else { 
		MLGetReal(stdlink, d);
		if (MLError(stdlink)) { 
			MLIF_error();
			*d = -997;
			MLClearError(stdlink);
		}
	}
}

/* TODO: Write callback receive for strings and vectors of strings */

/* ------------------------------------------------------------------
 *  Transfer vectors (callback)
 */

#ifdef CAPITAL
void W_STDCALL CBRECEIVENINT(int *remote_depth)
#else
#ifdef PREFIX
void cbreceivenint(int *remote_depth)
#else
void cbreceivenint_(int *remote_depth)
#endif
#endif
{
	cb_error = 0;
	cb_next = MLGetNext(stdlink);
	if (cb_next != MLTKFUNC) {
		MLIF_expected("IntegerList");
		*remote_depth = 1;
		cb_error = 1;
	}
	else { 
        MLGetInteger32List(stdlink, &local_data_int, &local_depth);
		*remote_depth = (int)local_depth;
		if (MLError(stdlink)) { 
			MLIF_error();
			*remote_depth = 1;
			cb_error = 2;
			MLClearError(stdlink);
		}
	}
}

#ifdef CAPITAL
void W_STDCALL CBCOPYNINT(int *remote_data_int)
#else
#ifdef PREFIX
void cbcopynint(int *remote_data_int)
#else
void cbcopynint_(int *remote_data_int)
#endif
#endif
{
	if (cb_error == 0) {
		memcpy((void *)remote_data_int, (void *)local_data_int,
		       sizeof(int)*local_depth);
        MLReleaseInteger32List(stdlink, local_data_int, local_depth);
	}
	else {
		*remote_data_int = -999+cb_error;
	}
}

#ifdef CAPITAL
void W_STDCALL CBRECEIVENDOUBLE(int *remote_depth)
#else
#ifdef PREFIX
void cbreceivendouble(int *remote_depth)
#else
void cbreceivendouble_(int *remote_depth)
#endif
#endif
{
	cb_error = 0;
	cb_next = MLGetNext(stdlink);
	if (cb_next != MLTKFUNC) {
		MLIF_expected("RealList");
		*remote_depth = 1;
		cb_error = 1;
	}
	else { 
        MLGetReal64List(stdlink, &local_data_double, &local_depth);
		*remote_depth = (int)local_depth;
		if (MLError(stdlink)) { 
			MLIF_error();
			*remote_depth = 1;
			cb_error = 2;
			MLClearError(stdlink);
		}
	}
}

#ifdef CAPITAL
void W_STDCALL CBCOPYNDOUBLE(double *remote_data_double)
#else
#ifdef PREFIX
void cbcopyndouble(double *remote_data_double)
#else
void cbcopyndouble_(double *remote_data_double)
#endif
#endif
{
	if (cb_error == 0) {
		memcpy((void *)remote_data_double, (void *)local_data_double,
		       sizeof(double)*local_depth);
        MLReleaseReal64List(stdlink, local_data_double, local_depth);
	}
	else {
		*remote_data_double = -999+cb_error;
	}
}

/* ------------------------------------------------------------------
 *  Transfer 2D arrays (callback)
 */

#ifdef CAPITAL
void W_STDCALL CBRECEIVENNINT(int *remote_dims)
#else
#ifdef PREFIX
void cbreceivennint(int *remote_dims)
#else
void cbreceivennint_(int *remote_dims)
#endif
#endif
{
	cb_error = 0;
	cb_next = MLGetNext(stdlink);
	if (cb_next != MLTKFUNC) {
		MLIF_expected("IntegerMatrix");
		*remote_dims = 1;
		*(remote_dims+1) = 1;
		cb_error = 1;
	}
	else { 
        MLGetInteger32Array(stdlink, &local_data_int, &local_dims,
		                  &local_heads, &local_depth);
		*remote_dims = (int)local_dims[0];
		*(remote_dims+1) = (int)local_dims[1];
		if (MLError(stdlink) || (local_depth != 2)) { 
			MLIF_error();
			*remote_dims = 1;
			*(remote_dims+1) = 1;
			cb_error = 2;
			MLClearError(stdlink);
		}
	}
} 

#ifdef CAPITAL
void W_STDCALL CBCOPYNNINT(int *remote_data_int)
#else
#ifdef PREFIX
void cbcopynnint(int *remote_data_int)
#else
void cbcopynnint_(int *remote_data_int)
#endif
#endif
{
	if (cb_error == 0) {
		memcpy((void *)remote_data_int, (void *)local_data_int,
		       sizeof(int)*local_dims[0]*local_dims[1]);
		MLReleaseInteger32Array(stdlink, local_data_int, local_dims,
		                     local_heads, local_depth);
	}
	else {
		*remote_data_int = -999+cb_error;
	}
}


#ifdef CAPITAL
void W_STDCALL CBRECEIVENNDOUBLE(int *remote_dims)
#else
#ifdef PREFIX
void cbreceivenndouble(int *remote_dims)
#else
void cbreceivenndouble_(int *remote_dims)
#endif
#endif
{
	cb_error = 0;
	cb_next = MLGetNext(stdlink);
	if (cb_next != MLTKFUNC) {
		MLIF_expected("RealMatrix");
		*remote_dims = 1;
		*(remote_dims+1) = 1;
		cb_error = 1;
	}
	else { 
        MLGetReal64Array(stdlink, &local_data_double, &local_dims,
		               &local_heads, &local_depth);
		*remote_dims = (int)local_dims[0];
		*(remote_dims+1) = (int)local_dims[1];
		if (MLError(stdlink) || (local_depth != 2)) { 
			MLIF_error();
			*remote_dims = 1;
			*(remote_dims+1) = 1;
			cb_error = 2;
			MLClearError(stdlink);
		}
	}
} 

#ifdef CAPITAL
void W_STDCALL CBCOPYNNDOUBLE(double *remote_data_double)
#else
#ifdef PREFIX
void cbcopynndouble(double *remote_data_double)
#else
void cbcopynndouble_(double *remote_data_double)
#endif
#endif
{
	if (cb_error == 0) {
		memcpy((void *)remote_data_double, (void *)local_data_double,
		       sizeof(int)*local_dims[0]*local_dims[1]);
		MLReleaseReal64Array(stdlink, local_data_double, local_dims,
		                  local_heads, local_depth);
	}
	else {
		*remote_data_double = -999+cb_error;
	}
}

/* ------------------------------------------------------------------
 *  Transfer 3D arrays (callback)
 */

#ifdef CAPITAL
void W_STDCALL CBRECEIVENNNINT(int *remote_dims)
#else
#ifdef PREFIX
void cbreceivennnint(int *remote_dims)
#else
void cbreceivennnint_(int *remote_dims)
#endif
#endif
{
	cb_error = 0;
	cb_next = MLGetNext(stdlink);
	if (cb_next != MLTKFUNC) {
		MLIF_expected("IntegerTensor of rank 3");
		*remote_dims = 1;
		*(remote_dims+1) = 1;
		*(remote_dims+2) = 1;
		cb_error = 1;
	}
	else { 
        MLGetInteger32Array(stdlink, &local_data_int, &local_dims,
		                  &local_heads, &local_depth);
		*remote_dims = (int)local_dims[0];
		*(remote_dims+1) = (int)local_dims[1];
		*(remote_dims+2) = (int)local_dims[2];
		if (MLError(stdlink) || (local_depth != 3)) { 
			MLIF_error();
			*remote_dims = 1;
			*(remote_dims+1) = 1;
			*(remote_dims+2) = 1;
			cb_error = 2;
			MLClearError(stdlink);
		}
	}
} 

#ifdef CAPITAL
void W_STDCALL CBCOPYNNNINT(int *remote_data_int)
#else
#ifdef PREFIX
void cbcopynnnint(int *remote_data_int)
#else
void cbcopynnnint_(int *remote_data_int)
#endif
#endif
{
	if (cb_error == 0) {
		memcpy((void *)remote_data_int, (void *)local_data_int,
		       sizeof(int)*local_dims[0]*local_dims[1]*local_dims[2]);
		MLReleaseInteger32Array(stdlink, local_data_int, local_dims,
		                     local_heads, local_depth);
	}
	else {
		*remote_data_int = -999+cb_error;
	}
}


#ifdef CAPITAL
void W_STDCALL CBRECEIVENNNDOUBLE(int *remote_dims)
#else
#ifdef PREFIX
void cbreceivennndouble(int *remote_dims)
#else
void cbreceivennndouble_(int *remote_dims)
#endif
#endif
{
	cb_error = 0;
	cb_next = MLGetNext(stdlink);
	if (cb_next != MLTKFUNC) {
		MLIF_expected("RealTensor of rank 3");
		*remote_dims = 1;
		*(remote_dims+1) = 1;
		*(remote_dims+2) = 1;
		cb_error = 1;
	}
	else { 
        MLGetReal64Array(stdlink, &local_data_double, &local_dims,
		               &local_heads, &local_depth);
		*remote_dims = (int)local_dims[0];
		*(remote_dims+1) = (int)local_dims[1];
		*(remote_dims+2) = (int)local_dims[2];
		if (MLError(stdlink) || (local_depth != 3)) { 
			MLIF_error();
			*remote_dims = 1;
			*(remote_dims+1) = 1;
			*(remote_dims+2) = 1;
			cb_error = 2;
			MLClearError(stdlink);
		}
	}
} 

#ifdef CAPITAL
void W_STDCALL CBCOPYNNNDOUBLE(double *remote_data_double)
#else
#ifdef PREFIX
void cbcopynnndouble(double *remote_data_double)
#else
void cbcopynnndouble_(double *remote_data_double)
#endif
#endif
{
	if (cb_error == 0) {
		memcpy((void *)remote_data_double, (void *)local_data_double,
		       sizeof(int)*local_dims[0]*local_dims[1]*local_dims[2]);
		MLReleaseReal64Array(stdlink, local_data_double, local_dims,
		                  local_heads, local_depth);
	}
	else {
		*remote_data_double = -999+cb_error;
	}
}

/* ------------------------------------------------------------------
 *  Transfer 4D arrays (callback)
 */

#ifdef CAPITAL
void W_STDCALL CBRECEIVENNNNINT(int *remote_dims)
#else
#ifdef PREFIX
void cbreceivennnnint(int *remote_dims)
#else
void cbreceivennnnint_(int *remote_dims)
#endif
#endif
{
	cb_error = 0;
	cb_next = MLGetNext(stdlink);
	if (cb_next != MLTKFUNC) {
		MLIF_expected("IntegerTensor of rank 4");
		*remote_dims = 1;
		*(remote_dims+1) = 1;
		*(remote_dims+2) = 1;
		*(remote_dims+3) = 1;
		cb_error = 1;
	}
	else { 
        MLGetInteger32Array(stdlink, &local_data_int, &local_dims,
		                  &local_heads, &local_depth);
		*remote_dims = (int)local_dims[0];
		*(remote_dims+1) = (int)local_dims[1];
		*(remote_dims+2) = (int)local_dims[2];
		*(remote_dims+3) = (int)local_dims[3];
		if (MLError(stdlink) || (local_depth != 4)) { 
			MLIF_error();
			*remote_dims = 1;
			*(remote_dims+1) = 1;
			*(remote_dims+2) = 1;
			*(remote_dims+3) = 1;
			cb_error = 2;
			MLClearError(stdlink);
		}
	}
} 

#ifdef CAPITAL
void W_STDCALL CBCOPYNNNNINT(int *remote_data_int)
#else
#ifdef PREFIX
void cbcopynnnnint(int *remote_data_int)
#else
void cbcopynnnnint_(int *remote_data_int)
#endif
#endif
{
	if (cb_error == 0) {
		memcpy((void *)remote_data_int, (void *)local_data_int,
		       sizeof(int)*local_dims[0]*local_dims[1]*local_dims[2]*local_dims[3]);
		MLReleaseInteger32Array(stdlink, local_data_int,
		                     local_dims, local_heads, local_depth);
	}
	else {
		*remote_data_int = -999+cb_error;
	}
}


#ifdef CAPITAL
void W_STDCALL CBRECEIVENNNNDOUBLE(int *remote_dims)
#else
#ifdef PREFIX
void cbreceivennnndouble(int *remote_dims)
#else
void cbreceivennnndouble_(int *remote_dims)
#endif
#endif
{
	cb_error = 0;
	cb_next = MLGetNext(stdlink);
	if (cb_next != MLTKFUNC) {
		MLIF_expected("RealTensor of rank 4");
		*remote_dims = 1;
		*(remote_dims+1) = 1;
		*(remote_dims+2) = 1;
		*(remote_dims+3) = 1;
		cb_error = 1;
	}
	else { 
        MLGetReal64Array(stdlink, &local_data_double, &local_dims,
		               &local_heads, &local_depth);
		*remote_dims = (int)local_dims[0];
		*(remote_dims+1) = (int)local_dims[1];
		*(remote_dims+2) = (int)local_dims[2];
		*(remote_dims+3) = (int)local_dims[3];
		if (MLError(stdlink) || (local_depth != 4)) { 
			MLIF_error();
			*remote_dims = 1;
			*(remote_dims+1) = 1;
			*(remote_dims+2) = 1;
			*(remote_dims+3) = 1;
			cb_error = 2;
			MLClearError(stdlink);
		}
	}
} 

#ifdef CAPITAL
void W_STDCALL CBCOPYNNNNDOUBLE(double *remote_data_double)
#else
#ifdef PREFIX
void cbcopynnnndouble(double *remote_data_double)
#else
void cbcopynnnndouble_(double *remote_data_double)
#endif
#endif
{
	if (cb_error == 0) {
		memcpy((void *)remote_data_double, (void *)local_data_double,
		       sizeof(int)*local_dims[0]*local_dims[1]*local_dims[2]*local_dims[3]);
		MLReleaseReal64Array(stdlink, local_data_double, local_dims,
		                  local_heads, local_depth);
	}
	else {
		*remote_data_double = -999+cb_error;
	}
}

